SystemOrganization addCategory: #CS528su14!
SystemOrganization addCategory: 'CS528su14-Payroll'!
SystemOrganization addCategory: 'CS528su14-ValueWithHistory'!

TestCase subclass: #EmployeeTestCase
	instanceVariableNames: 'susan day1 normalMonth aLittleOvertime'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!EmployeeTestCase methodsFor: 'Running' stamp: 'dd 1/11/2005 19:36'!
setUp
	day1 := Date 
				newDay: 11
				month: 7
				year: 2001.
	susan := Employee new named: 'Susan'.
	normalMonth := #(#(40 0) #(40 0) #(40 0) #(40 0)).
	aLittleOvertime := #(#(40 0) #(50 0) #(40 0) #(40 0))
			! !

!EmployeeTestCase methodsFor: 'Testing' stamp: 'MattMomont 6/22/2014 17:29'!
testChangeSalary
	susan changeSalaryFor: day1 to: 20.
	self assert: (susan salaryAt: day1)= 20.
	day1 addDays: 7.
	susan changeSalaryFor: day1 to: 25.
	self assert: (susan salaryAt: day1) = 25
			! !

TestCase subclass: #PayrollSystemHistoryTestCase
	instanceVariableNames: 'myPayroll day1 susan normalMonth day20 endOfMonth'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!PayrollSystemHistoryTestCase methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 12:16'!
postTimeCards: collection to: anEmployee on: aDate
	| timecardDay |
	timecardDay := aDate.
	collection do: 
			[:each | 
			anEmployee 
				postTimeCardFor: timecardDay
				hoursWorked: each first
				vacation: each last.
			timecardDay := timecardDay addDays: 7]
			! !

!PayrollSystemHistoryTestCase methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 17:37'!
setUp
	myPayroll := PayrollSystem new name: 'ACME'.
	day1 := Date newDay: 11 month: 7 year: 2001.
	day20 := day1 addDays: 19.
	endOfMonth := day20 addDays: 8.
	susan := Employee new named: 'Susan'.
	susan changeSalaryFor: day1 to: 20.
	normalMonth := #(#(40 0) #(40 0) #(40 0) #(40 0)).

		! !

!PayrollSystemHistoryTestCase methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 17:37'!
testEarnings
	myPayroll addEmployee: susan.
	self postTimeCards: normalMonth to: susan on: day1.
	self assert: (susan earnedAt: endOfMonth) = 3200! !

!PayrollSystemHistoryTestCase methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 17:43'!
testLateTimeCard
	| aPreviousWeek |
	aPreviousWeek := #(#(40 0)).
	myPayroll addEmployee: susan.
	susan changeSalaryFor: day20 to: 50.
	self postTimeCards: aPreviousWeek to: susan on: day1.
	self assert: (susan earnedAt: endOfMonth) = 800! !

!PayrollSystemHistoryTestCase methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 18:48'!
testMidMonthRaise
	| result |
	myPayroll addEmployee: susan.
	susan changeSalaryFor: day20 to: 50.
	self postTimeCards: normalMonth to: susan on: day1.
	self assert: (susan earnedAt: endOfMonth) = 4400.

	result := String new writeStream.
	susan salaryHistoryOn: result.
	self assert: result contents = '31 December -001 : $0 11 July 2001 : $20 30 July 2001 : $50 '.
! !

!PayrollSystemHistoryTestCase methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 18:40'!
testPaycheck
	| paycheck |
	myPayroll addEmployee: susan.
	self postTimeCards: normalMonth to: susan on: day1.
	self assert: (susan paidAt: endOfMonth) = 0.
	self assert: (susan earnedAt: endOfMonth) = 3200.
	
	paycheck := susan makePaycheckAt: (endOfMonth addDays: 10).
	
	"Total earned - withholdings"
	self assert: (paycheck amountPaid) = 2377.6.
	
	"Susan should have been paid"
	susan postTransaction: paycheck.
	self assert: (susan paidAt: (endOfMonth addDays: 11)) = 2377.6.
! !

!PayrollSystemHistoryTestCase methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 18:25'!
testWithholdings
	| paycheck |
	myPayroll addEmployee: susan.
	susan incrementWithholding: 50 at: day20. 
	self postTimeCards: normalMonth to: susan on: day1.
	self assert: (susan paidAt: endOfMonth) = 0.
	self assert: (susan earnedAt: endOfMonth) = 3200.
	
	paycheck := susan makePaycheckAt: (endOfMonth addDays: 10).
	
	"Total earned - withholdings <-- More withholdings this time"
	self assert: (paycheck amountPaid) = 2340.45.
	
	"Susan should have been paid"
	susan postTransaction: paycheck.
	self assert: (susan paidAt: (endOfMonth addDays: 11)) =  2340.45.
! !

TestCase subclass: #PayrollSystemTestCase
	instanceVariableNames: 'myPayroll day1 susan george normalMonth aLittleOvertime endOfMonth day20'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

PayrollSystemTestCase subclass: #PayrollInsuraceChangeTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!PayrollInsuraceChangeTest methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 20:24'!
testInsuranceWithheld
	| paycheck check |
	
	myPayroll addEmployee: susan.

	"Month #1 - insurance rate of $40 "
	self postTimeCards: normalMonth to: susan.
	
	susan changeInsuranceFor: day20 to: 40.
	check := String new writeStream.
	paycheck := susan makePaycheckAt: endOfMonth.
	paycheck printOnCheckStream: check.
	self assert: (check contents findString: 'insurance: $40') > 0.
	
	susan postTransaction: paycheck.
	self assert: (susan earnedAt: endOfMonth) = 3200.
	self assert: (susan paidAt: endOfMonth) = 2347.88. "Earned - witholdings - insurance"

	"Month #2 - insurance rate of $50"
	self postTimeCards: normalMonth to: susan on: endOfMonth.
	susan changeInsuranceFor: (endOfMonth addDays: 5 ) to: 50.
	paycheck := susan makePaycheckAt: (endOfMonth addDays: 28).
	check := String new writeStream.
	paycheck printOnCheckStream: check.
	self assert: (check contents findString: 'insurance: $50') > 0.
	self assert: (check contents findString: 'insurance: $90') > 0. ! !

!PayrollSystemTestCase methodsFor: 'private' stamp: ''!
postTimeCards: collection to: anEmployee 
	| timecardDay |
	timecardDay := day1.
	collection do: 
			[:each | 
			anEmployee 
				postTimeCardFor: timecardDay
				hoursWorked: each first
				vacation: each last.
			timecardDay := timecardDay addDays: 7]
			! !

!PayrollSystemTestCase methodsFor: 'private' stamp: 'MattMomont 6/22/2014 19:58'!
postTimeCards: collection to: anEmployee on: aDate
	| timecardDay |
	timecardDay := aDate.
	collection do: 
			[:each | 
			anEmployee 
				postTimeCardFor: timecardDay
				hoursWorked: each first
				vacation: each last.
			timecardDay := timecardDay addDays: 7]
			! !

!PayrollSystemTestCase methodsFor: 'Running' stamp: 'MattMomont 6/22/2014 19:04'!
setUp
	myPayroll := PayrollSystem new name: 'ACME'.
	day1 := Date newDay: 11 month: 7 year: 2001.
	day20 := day1 addDays: 19.
	endOfMonth := day1 addDays: 27. 
	susan := Employee new named: 'Susan'.
	susan changeSalaryFor: day1 to: 20.
	george := Employee new named: 'George'.
	george changeSalaryFor: day1 to: 12.
	normalMonth := #(#(40 0) #(40 0) #(40 0) #(40 0)).
	aLittleOvertime := #(#(40 0) #(50 0) #(40 0) #(40 0)).
		! !

!PayrollSystemTestCase methodsFor: 'Testing' stamp: ''!
testAddEmployee
	myPayroll addEmployee: susan.
	self assert: (myPayroll employeeNames includes: 'Susan').
	myPayroll addEmployee: george.
	self assert: (myPayroll employeeNamed: 'Susan') = susan
			! !

!PayrollSystemTestCase methodsFor: 'Testing' stamp: 'MattMomont 6/22/2014 17:44'!
testEarnings
	myPayroll
		addEmployee: susan;
		addEmployee: george.
	self postTimeCards: normalMonth to: susan.
	self postTimeCards: aLittleOvertime to: george.
	self assert: (susan earnedAt: endOfMonth) = 3200.
	self assert: (george earnedAt: endOfMonth) = 2100! !

!PayrollSystemTestCase methodsFor: 'Testing' stamp: 'MattMomont 6/21/2014 20:08'!
testSerialization
	| restoredPayroll |
	myPayroll := PayrollSystem new name: 'BackMeUp'.
	myPayroll addEmployee: susan.
	myPayroll save.
	
	restoredPayroll := myPayroll restore.
	self assert: restoredPayroll name = 'BackMeUp'.
	self assert: (restoredPayroll employeeNames at: 1) = 'Susan'.! !

PayrollSystemTestCase subclass: #PayrollVacationTest
	instanceVariableNames: 'aVacationMonth'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!PayrollVacationTest methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/20/2014 21:50'!
setUp
	super setUp.
	aVacationMonth := #(#(40 0) #(40 0) #(40 0) #(32 8)).! !

!PayrollVacationTest methodsFor: 'Testing' stamp: 'MattMomont 6/22/2014 18:35'!
testVacation
	myPayroll addEmployee: susan.	"4 weeks x 40 hrs = 160 / 25 = 32/5"
	self postTimeCards: normalMonth to: susan.
	self assert: (susan earnedAt: endOfMonth) = 3200.
	self assert: (susan vacationAt: endOfMonth) = (32 / 5).	"	
	#(#(40 0) #(40 0) #(40 0) #(32 8))
	Added on to the timecard above
	(152 + 160) / 25 - 8 = 4.48 = 112/25
 	"
	self postTimeCards: aVacationMonth to: susan.
	self assert: (susan earnedAt: endOfMonth) = 6240.
	self assert: (susan vacationAt: endOfMonth) = (112 / 25)! !

TestCase subclass: #SortedSequenceTestCase
	instanceVariableNames: 'sequence'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!SortedSequenceTestCase methodsFor: 'Running' stamp: ''!
setUp
	sequence := SortedSequence new.
	sequence add: 1; add: 80; add: 500; add: 501
			! !

!SortedSequenceTestCase methodsFor: 'Testing' stamp: ''!
testIndex
	"(SortedSequenceTestCase selector: #testIndex) run"

	self assert: (sequence indexOfStartOfIntervalContaining: 1) = 1.
	self assert: (sequence indexOfStartOfIntervalContaining: 79) = 1.
	self assert: (sequence indexOfStartOfIntervalContaining: 80) = 2.
	self assert: (sequence indexOfStartOfIntervalContaining: 501)	= 4
			! !

TestCase subclass: #ValueWithHistoryTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-ValueWithHistory'!

!ValueWithHistoryTest methodsFor: 'testing' stamp: 'ChaseGeigle 6/19/2014 20:16'!
testInvalid
	| history |
	history := ValueWithHistory new.
	history starting: (Date year: 2014 month: 06 day: 19) become: 1.
	self should: [ history at: (Date year: 2013 month: 06 day: 19) ] raise: Error! !

!ValueWithHistoryTest methodsFor: 'testing' stamp: 'ChaseGeigle 6/19/2014 20:14'!
testValid
	| history day3 day2 day1 day4 |
	day3 := '12/12/2007' asDate.
	day2 := day3 subtractDays: 3.
	day1 := day2 subtractDays: 3.
	day4 :=  day3 addDays: 4.
	(history := ValueWithHistory new) starting: day3 become: 3.
	history starting: day1 become: 1.
	self should: [(history at: day2) = 1].
	history at: day2 add: 1.
	self should: [(history at: day3) = 4].
	history at: day4 add: 2.
	self should: [(history at: day4) = 6].
! !

SortedCollection subclass: #SortedSequence
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!SortedSequence commentStamp: '<historical>' prior: 0!
I add one method to SortedCollection.  It would be just as easy to add the method
directly to the class instead of making a subclass, but this makes it easier to see
the new method.

The method is indexOfStartOfIntervalContaining: anElement
It returns the index of anElement or, if it is not present, the index of the
largest element smaller than it.  It is useful when you want to have a
table that is sorted by keys and you want to look up in the table to
find the nearest match to a key.!

!SortedSequence methodsFor: 'accessing' stamp: ''!
indexOfStartOfIntervalContaining: anElement 
	"Return the index of anElement or, if it is not present, of the index of the
	largest element smaller than it."

	self isEmpty ifTrue: [^0].
	^(self indexForInserting: anElement) - firstIndex
			! !

Object subclass: #Employee
	instanceVariableNames: 'name transactions salary earned paid withholding accruedVacation taxRule insurance totalInsurance withholders'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!Employee commentStamp: '<historical>' prior: 0!
I represent an employee in a payroll system.  Thus, I keep track of the number of hours worked,
the wages paid, the vacation accrued, the taxes withheld, and so on.  My values change only
when transactions are posted to me.  Transactions are subclasses of EmployeeTransaction.

Variables
	name 				<String>
	transactions 			<Collection of Transactions, sorted by date>
	salary 				<ValueWithHistory>
	earned  				<ValueWithHistory>
	paid  				<ValueWithHistory>
	withholding  			<ValueWithHistory>
	accruedVacation  	<ValueWithHistory>
	taxRule 				<TaxRule>!

!Employee class methodsFor: 'class initialization' stamp: 'MattMomont 6/22/2014 20:52'!
initialize
	"Employee initialize"

! !

!Employee methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 20:03'!
changeInsuranceFor: aDate to: anInsuranceRate 
	self postTransaction: (InsuranceChange new date: aDate employee: self rate: anInsuranceRate)! !

!Employee methodsFor: 'actions' stamp: 'ChaseGeigle 6/7/2014 00:01'!
changeSalaryFor: aDate to: dollarPerHour
	self postTransaction: (SalaryChange new date: aDate employee: self salary: dollarPerHour)! !

!Employee methodsFor: 'accessing' stamp: 'MattMomont 6/22/2014 17:33'!
earnedAt: aDate
	^ earned at: aDate.! !

!Employee methodsFor: 'private' stamp: 'ChaseGeigle 6/7/2014 00:02'!
federalTaxesOn: money assumingTotalIs: estimatedEarnings
	^ money * (taxRule rateFor: estimatedEarnings)! !

!Employee methodsFor: 'posting' stamp: 'MattMomont 6/22/2014 17:34'!
incrementEarned: anAmount at: aDate
	earned at:aDate add:anAmount! !

!Employee methodsFor: 'posting' stamp: 'MattMomont 6/22/2014 20:21'!
incrementInsurance: anAmount at: aDate
	totalInsurance at: aDate add: anAmount! !

!Employee methodsFor: 'posting' stamp: 'MattMomont 6/22/2014 17:54'!
incrementPaid: anAmount at: aDate
	"Update the amount paid at a given point in time"
	paid at: aDate add: anAmount! !

!Employee methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 18:34'!
incrementVacation: aNumber at: aDate 
	accruedVacation at: aDate add: aNumber.! !

!Employee methodsFor: 'posting' stamp: 'MattMomont 6/22/2014 18:22'!
incrementWithholding: anAmount at: aDate
	withholding at: aDate add: anAmount! !

!Employee methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 19:09'!
insurance: anAmount at: aDate 
	insurance starting: aDate become: anAmount.! !

!Employee methodsFor: 'actions' stamp: 'MattMomont 6/22/2014 21:23'!
makePaycheckAt: aDate
	"Create a paycheck for the amount owed to the employee. It hasn't 
	been posted."

	| amountPaid amountWithheld |
	amountPaid := (earned at: aDate) - (paid at: aDate) - (withholding at: aDate) - (insurance at: aDate).
	amountWithheld := TaxWithholder new withholdingOn: amountPaid for: self at: aDate.
	amountPaid := amountPaid - amountWithheld.
	
	"TODO: for each withholder, edit the pay check and return"
	
	^ Paycheck
		date: aDate
		employee: self
		paid: amountPaid
		withheld: amountWithheld
		insurance: (insurance at: aDate)
		totalPaid: (paid at: aDate) + amountPaid
		totalWithheld: (withholding at: aDate) + amountWithheld 
		totalInsurance: (totalInsurance at: aDate) + (insurance at: aDate)! !

!Employee methodsFor: 'accessing' stamp: 'ChaseGeigle 6/7/2014 00:03'!
name
	^ name! !

!Employee methodsFor: 'initialize-release' stamp: 'MattMomont 6/22/2014 20:56'!
named: aString
	name := aString.
	transactions := OrderedCollection new.
	salary := ValueWithHistory zero.
	earned := ValueWithHistory zero.
	paid := ValueWithHistory zero.
	withholding := ValueWithHistory zero.
	accruedVacation := ValueWithHistory zero.
	insurance := ValueWithHistory zero.
	totalInsurance := ValueWithHistory zero.
	withholders := Dictionary new.
	withholders at: #taxes put: (TaxWithholder new).! !

!Employee methodsFor: 'accessing' stamp: 'MattMomont 6/22/2014 17:52'!
paidAt: aDate
	^ paid at: aDate.! !

!Employee methodsFor: 'actions' stamp: 'ChaseGeigle 6/7/2014 00:01'!
postTimeCardFor: aDate hoursWorked: hoursW vacation: hoursV
	self
		postTransaction:
			(Timecard
				date: aDate
				employee: self
				worked: hoursW
				vacation: hoursV)! !

!Employee methodsFor: 'posting' stamp: 'ChaseGeigle 6/7/2014 00:03'!
postTransaction: aTransaction
	"Public interface to record and process (i.e. post) a transaction."

	aTransaction postTo: self.
	transactions add: aTransaction! !

!Employee methodsFor: 'accessing' stamp: 'MattMomont 6/22/2014 17:33'!
salary: anAmount at: aDate
	"Set a new salary on a date"
	salary starting: aDate become: anAmount.! !

!Employee methodsFor: 'accessing' stamp: 'MattMomont 6/22/2014 17:29'!
salaryAt: aDate
	^ salary at: aDate.! !

!Employee methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 18:45'!
salaryHistoryOn: aStream
	salary printOn: aStream.! !

!Employee methodsFor: 'private' stamp: 'MattMomont 6/22/2014 18:05'!
taxesOn: money at: date
	"Return the taxes to be withheld from the money earned.
	This is just a guess - I didn't look up the tax rules."

	| socialSecurity federalTaxes stateTaxes estimatedEarnings |
	socialSecurity := (0 max: (money min: 60600 - (earned at: date))) * 0.077.
	estimatedEarnings := ((earned at: date) + money) * date day / date daysInYear.
	federalTaxes := self federalTaxesOn: money assumingTotalIs: estimatedEarnings.
	stateTaxes := money * 0.03.	"Illinois has a flat tax."
	^ socialSecurity + federalTaxes + stateTaxes! !

!Employee methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 18:35'!
vacationAt: aDate
	^accruedVacation at: aDate.! !

Object subclass: #EmployeeTransaction
	instanceVariableNames: 'date employee'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!EmployeeTransaction commentStamp: '<historical>' prior: 0!
I'm an abstract class of transactions that can be posted to an Employee.

My most important method is postTo:, which is implemented by subclasses
to send a message to the Employee to post the transaction.  The Employee
has all the information, so it has to be the one to implement posting, but the
postTo: method implements a standard interface so that clients of the
payroll system can always post a transaction by sending the postTransaction:
message to an Employee.

Variables
	date		<Date>   Date at which transaction occured.
	employee	<Employee>

!

!EmployeeTransaction methodsFor: 'accessing' stamp: ''!
date
	^date
			! !

!EmployeeTransaction methodsFor: 'posting' stamp: ''!
post
	employee postTransaction: self
			! !

!EmployeeTransaction methodsFor: 'posting' stamp: ''!
postTo: anEmployee
	"double dispatching"
	self subclassResponsibility
			! !

EmployeeTransaction subclass: #InsuranceChange
	instanceVariableNames: 'newRate'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!InsuranceChange commentStamp: '<historical>' prior: 0!
An InsuranceChange is a transaction that allows you to set the current insurance rate to be withheld from an Employee's paycheck.
!

!InsuranceChange methodsFor: 'initialize' stamp: 'MattMomont 6/22/2014 19:07'!
date: aDate employee: anEmployee rate: aRate 
	date := aDate.
	employee := anEmployee.
	newRate := aRate! !

!InsuranceChange methodsFor: 'posting' stamp: 'MattMomont 6/22/2014 19:08'!
postTo: anEmployee 
	anEmployee insurance: newRate at: date.
			! !

EmployeeTransaction subclass: #Paycheck
	instanceVariableNames: 'amountPaid taxes totalPaid totalTaxes insurance totalInsurance'
	classVariableNames: 'AmountFormat DateFormat'
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!Paycheck commentStamp: '<historical>' prior: 0!
Transaction representing a payment to an employee.!

!Paycheck class methodsFor: 'instance creation' stamp: 'MattMomont 6/22/2014 19:53'!
date: aDate employee: anEmployee paid: amount1 withheld: amount2 insurance: aRate totalPaid: amount3 totalWithheld: amount4 totalInsurance: amount5
	^ self new
		date: aDate
		employee: anEmployee
		paid: amount1
		withheld: amount2
		insurance: aRate
		totalPaid: amount3
		totalWithheld: amount4
		totalInsurance: amount5! !

!Paycheck methodsFor: 'accessing' stamp: ''!
amountPaid
	^amountPaid
			! !

!Paycheck methodsFor: 'initialize-release' stamp: 'MattMomont 6/22/2014 19:51'!
date: aDate employee: anEmployee paid: amount1 withheld: amount2 insurance: amount3 totalPaid: amount4 totalWithheld: amount5 totalInsurance: amount6
	date := aDate.
	employee := anEmployee.
	amountPaid := amount1.
	taxes := amount2.
	insurance := amount3.
	totalPaid := amount4.
	totalTaxes := amount5.
	totalInsurance := amount6.
			! !

!Paycheck methodsFor: 'posting' stamp: 'MattMomont 6/22/2014 19:26'!
postTo: anEmployee
	anEmployee incrementPaid: amountPaid at: date.
	anEmployee incrementWithholding: taxes at: date.
	anEmployee incrementInsurance: insurance at: date.! !

!Paycheck methodsFor: 'printing' stamp: 'MattMomont 6/22/2014 20:24'!
printOnCheckStream: aStream
	"This method is supposed to print information on a check that is on the printer, and
	aStream is the printer.  Unfortunately, the whole method is bogus, because we don't
	know what the printer or the check are like.  I'm assuming that the check has two-parts.
	One is the actual check, and we need to print the employee's name and the amount.
	The second is the check stub, which will say how much is being paid and how much
 	is being withheld. I'm also assuming that the check is preprinted with the name of the
	company, and everything else that is usual for a check."

	aStream cr; cr.
	aStream next: 40 put: (Character space).
	"DateFormat print: date on: aStream."
	date printOn: aStream.
	aStream cr.
	aStream nextPutAll: ((String new: 40 withAll: Character space) copyReplaceFrom: 1 to: employee name size with: employee name).
	"AmountFormat print: amountPaid on: aStream."
	amountPaid printOn: aStream.
	aStream cr; cr; cr; cr.
	aStream nextPutAll: 'this month  '.
	"AmountFormat print: amountPaid on: aStream.
	AmountFormat print: taxes on: aStream."
	aStream nextPutAll: ' amountPaid: $'.
	amountPaid printOn: aStream.
	aStream nextPutAll: ' taxes: $'.
	taxes printOn: aStream.
	aStream nextPutAll: ' insurance: $'.
	insurance printOn: aStream.
	aStream cr.
	aStream nextPutAll: 'year to date '.
	"AmountFormat print: totalPaid on: aStream.
	AmountFormat print: totalTaxes on: aStream."
	aStream nextPutAll: ' amountPaid: $'.
	totalPaid printOn: aStream.
	aStream nextPutAll: ' taxes: $'.
	totalTaxes printOn: aStream.
	aStream nextPutAll: ' insurance: $'.
	totalInsurance printOn: aStream.	
	aStream cr; cr.
	! !

!Paycheck methodsFor: 'accessing' stamp: ''!
taxes
	^taxes
			! !

EmployeeTransaction subclass: #SalaryChange
	instanceVariableNames: 'newSalary'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!SalaryChange commentStamp: '<historical>' prior: 0!
Transaction representing a change in salary, usually a raise.!

!SalaryChange methodsFor: 'initialize-release' stamp: ''!
date: aDate employee: anEmployee salary: dollarPerHour
	date := aDate.
	employee := anEmployee.
	newSalary := dollarPerHour
			! !

!SalaryChange methodsFor: 'accessing' stamp: ''!
newSalary
	^newSalary
			! !

!SalaryChange methodsFor: 'posting' stamp: 'MattMomont 6/22/2014 17:33'!
postTo: anEmployee
	anEmployee salary: newSalary at: date.
			! !

EmployeeTransaction subclass: #Timecard
	instanceVariableNames: 'hoursWorked hoursVacation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!Timecard commentStamp: '<historical>' prior: 0!
Transaction representing the work that the employee has done.!

!Timecard class methodsFor: 'instance creation' stamp: 'ChaseGeigle 6/7/2014 00:08'!
date: aDate employee: anEmployee worked: hoursW vacation: hoursV
	^ self new
		date: aDate
		employee: anEmployee
		worked: hoursW
		vacation: hoursV! !

!Timecard methodsFor: 'initialize-release' stamp: ''!
date: aDate employee: anEmployee worked: hoursW vacation: hoursV
	date := aDate.
	employee := anEmployee.
	hoursWorked := hoursW.
	hoursVacation := hoursV
			! !

!Timecard methodsFor: 'accessing' stamp: ''!
hoursVacation
	^hoursVacation
			! !

!Timecard methodsFor: 'accessing' stamp: ''!
hoursWorked
	^hoursWorked
			! !

!Timecard methodsFor: 'posting' stamp: 'MattMomont 6/22/2014 18:34'!
postTo: anEmployee
	| money overtime nonovertime vacation |
	overtime := hoursWorked - 40 max: 0.
	nonovertime := hoursWorked min: 40.
	money := (overtime * 1.5 + nonovertime) * (anEmployee salaryAt: date).
	anEmployee incrementEarned: money at: date.
	vacation := hoursWorked / 25 - hoursVacation.
	anEmployee incrementVacation: vacation at: date! !

Object subclass: #PayrollSystem
	instanceVariableNames: 'employees paychecks companyName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!PayrollSystem commentStamp: '<historical>' prior: 0!
I am the top-level of the payroll system.  I keep track of the employees
and can "run the payroll", i.e. write paychecks for the employees.  To
run the payroll, send the messages
	makePaychecksAt: paydate	- to create the paycheck transactions
	printPaychecksOn: aStream	- to print them out 
	postPaychecks				- to record them in employees

Note that the employees are not marked with having been paid until
the paychecks are posted, so just creating the transactions or even
printing them is not enough.

Variables
	employees   <Dictionary from String to Employee>
	paychecks 	<Collection of Paycheck>	holds paychecks that have been created until they have been posted!

!PayrollSystem methodsFor: 'accessing' stamp: ''!
addEmployee: anEmployee
	employees at: anEmployee name put: anEmployee
			! !

!PayrollSystem methodsFor: 'accessing' stamp: ''!
employeeNamed: aName
	^employees at: aName
			! !

!PayrollSystem methodsFor: 'accessing' stamp: ''!
employeeNames
	^employees keys asOrderedCollection
			! !

!PayrollSystem methodsFor: 'initialize-release' stamp: ''!
initialize
	employees := Dictionary new.
			! !

!PayrollSystem methodsFor: 'actions' stamp: ''!
makePaychecksAt: aDate 
	paychecks := employees collect: [:eachEmployee | eachEmployee makePaycheckAt: aDate]
			! !

!PayrollSystem methodsFor: 'accessing' stamp: ''!
name
	^companyName
			! !

!PayrollSystem methodsFor: 'accessing' stamp: ''!
name: aString
	companyName := aString
			! !

!PayrollSystem methodsFor: 'actions' stamp: ''!
paycheckString
	| aStream |
	aStream := WriteStream on: (String new: 20).
	self printPaychecksOn: aStream.
	^aStream contents
			! !

!PayrollSystem methodsFor: 'actions' stamp: ''!
postPaychecks
	paychecks do: [:each | each post].
	paychecks := #()
			! !

!PayrollSystem methodsFor: 'actions' stamp: ''!
printPaychecksOn: aStream 
	paychecks do: [:each | each printOnCheckStream: aStream]
			! !

!PayrollSystem methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/21/2014 19:59'!
restore
	"Serialize Payroll to disk"
	^FLMaterializer materializeFromFileNamed: 'payroll-backup.fl'.
! !

!PayrollSystem methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/21/2014 19:55'!
save
	"Serialize Payroll to disk"
	FLSerializer serialize: self toFileNamed: 'payroll-backup.fl'.

! !

Object subclass: #TaxRule
	instanceVariableNames: 'breakPoints rates'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!TaxRule commentStamp: '<historical>' prior: 0!
I represent a tax table.  Given a yearly taxable income, I can give the tax rate for it.
I store the tax table as a sequence of amounts, and the tax rate for that amount.
For example, in 1994 the tax rate for a single person (schedule X) was
           $0		15%
  $22750	28%
  $55100	31%
$115000	36%
$250000	39.6%

breakpoints  	<SortedCollection of: integer>
rates	 	<OrderedCollection of: integer>!

!TaxRule methodsFor: 'accessing' stamp: ''!
breakPoint: aNumber rate: anotherNumber 
	"Add an entry to the tax table"

	| index |
	index := breakPoints indexOfStartOfIntervalContaining: aNumber.
	breakPoints add: aNumber.
	rates add: anotherNumber beforeIndex: index + 1
			! !

!TaxRule methodsFor: 'initialize-release' stamp: ''!
initialize
	breakPoints := SortedSequence new.
	rates := OrderedCollection new.
			! !

!TaxRule methodsFor: 'accessing' stamp: ''!
rateFor: aNumber 
	"Return the tax rate for someone whose taxable income is 
	aNumber dollars."

	^rates at: (breakPoints indexOfStartOfIntervalContaining: aNumber)
			! !

Object subclass: #ValueWithHistory
	instanceVariableNames: 'date value'
	classVariableNames: 'EarliestDate'
	poolDictionaries: ''
	category: 'CS528su14-ValueWithHistory'!

!ValueWithHistory commentStamp: '<historical>' prior: 0!
I represent a (probably numerical) value that changes over time.  I can answer my value at any point in
time (usually a date).  I can update my value at any point in time, and also add a number to my
value from any point in time on into the future.  My value does not change continuously, but changes
at discrete points in time.

Instance Variables:
	date	<SortedCollection of: Date>
	value	<OrderedCollection of: Number>
	The i'th element of value matches the i'th element of date.  The date collection indicates when the value takes on the next element of the value collection.
!

!ValueWithHistory class methodsFor: 'class initialization' stamp: 'ChaseGeigle 6/19/2014 20:04'!
initialize
	EarliestDate := Date newDay: 0 year: 0.
! !

!ValueWithHistory class methodsFor: 'instance creation' stamp: 'm3r 6/27/2008 02:06'!
zero
	^ self new starting: EarliestDate become: 0
! !

!ValueWithHistory methodsFor: 'accessing' stamp: 'ChaseGeigle 6/19/2014 20:09'!
at: aDate
	"Return value at aDate"

	| index |
	index := date indexOfStartOfIntervalContaining: aDate.
	index = 0
		ifTrue: [ self error: 'date is too early' ].
	^ value at: index! !

!ValueWithHistory methodsFor: 'accessing' stamp: 'ChaseGeigle 6/19/2014 20:09'!
at: aDate add: amount
	"Add amount to the value starting at aDate."

	| start |
	start := self indexForAccessing: aDate.
	start to: value size do: [ :each | value at: each put: (value at: each) + amount ]! !

!ValueWithHistory methodsFor: 'accessing' stamp: 'ChaseGeigle 6/19/2014 20:09'!
indexForAccessing: aDate
	"Create index of slot for aDate, creating it if necessary."

	| index |
	(index := date indexOfStartOfIntervalContaining: aDate) = 0
		ifTrue: [ 
			date add: aDate.
			value addFirst: 0.
			^ 1 ].
	(date at: index) = aDate
		ifTrue: [ ^ index ].
	date add: aDate.
	value add: (value at: index) beforeIndex: index + 1.
	^ index + 1! !

!ValueWithHistory methodsFor: 'initialize-release' stamp: 'ChaseGeigle 6/19/2014 20:09'!
initialize
	date := SortedSequence new.
	value := OrderedCollection new! !

!ValueWithHistory methodsFor: 'printing' stamp: 'MattMomont 6/22/2014 18:48'!
printOn: aStream
	"aStream nextPut: $'."
	date
		with: value
		do: [ :nextDate :nextValue | 
			aStream
				print: nextDate;
				nextPutAll: ' : $';
				print: nextValue;
				nextPutAll: ' '].
	"aStream nextPut: $'."! !

!ValueWithHistory methodsFor: 'accessing' stamp: 'ChaseGeigle 6/19/2014 20:10'!
starting: aDate become: anAmount
	| index |
	date add: aDate.
	index := date indexOfStartOfIntervalContaining: aDate.
	value add: anAmount beforeIndex: index! !

Object subclass: #Withholder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!Withholder commentStamp: '<historical>' prior: 0!
A Withholder is somebody who takes money from each paycheck.
!

Withholder subclass: #TaxWithholder
	instanceVariableNames: 'rule'
	classVariableNames: 'HeadTaxRate MarriedSeparateTaxRate MarriedTaxRate SingleTaxRate'
	poolDictionaries: ''
	category: 'CS528su14-Payroll'!

!TaxWithholder commentStamp: '<historical>' prior: 0!
A TaxWithholder is the worst kind of withholder. Free money for the government.
!

!TaxWithholder methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 21:13'!
federalTaxesOn: money assumingTotalIs: estimatedEarnings
	^ money * (rule rateFor: estimatedEarnings)! !

!TaxWithholder methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 21:22'!
initialize
	SingleTaxRate := TaxRule new.
	SingleTaxRate
		breakPoint: 0 rate: 0.15;
		breakPoint: 22750 rate: 0.28;
		breakPoint: 55100 rate: 0.31;
		breakPoint: 115000 rate: 0.36;
		breakPoint: 250000 rate: 0.396.
		
	MarriedTaxRate := TaxRule new.
	MarriedTaxRate
		breakPoint: 0 rate: 0.15;
		breakPoint: 38000 rate: 0.28;
		breakPoint: 91850 rate: 0.31;
		breakPoint: 140000 rate: 0.36;
		breakPoint: 250000 rate: 0.396.
		
	MarriedSeparateTaxRate := TaxRule new.
	MarriedSeparateTaxRate
		breakPoint: 0 rate: 0.15;
		breakPoint: 19500 rate: 0.28;
		breakPoint: 45925 rate: 0.31;
		breakPoint: 70000 rate: 0.36;
		breakPoint: 125000 rate: 0.396.
		
	HeadTaxRate := TaxRule new.
	HeadTaxRate
		breakPoint: 0 rate: 0.15;
		breakPoint: 30500 rate: 0.28;
		breakPoint: 78700 rate: 0.31;
		breakPoint: 127500 rate: 0.36;
		breakPoint: 250000 rate: 0.396.
		
	rule := SingleTaxRate.! !

!TaxWithholder methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 20:48'!
rule: aTaxRule
	rule := aTaxRule.! !

!TaxWithholder methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 21:15'!
withholdingOn: money for: anEmployee at: date
	"Return the taxes to be withheld from the money earned.
	This is just a guess - I didn't look up the tax rules."

	| socialSecurity federalTaxes stateTaxes estimatedEarnings |
	socialSecurity := (0 max: (money min: 60600 - (anEmployee earnedAt: date))) * 0.077.
	estimatedEarnings := ((anEmployee earnedAt: date) + money) * date day / date daysInYear.
	federalTaxes := self federalTaxesOn: money assumingTotalIs: estimatedEarnings.
	stateTaxes := money * 0.03.	"Illinois has a flat tax."
	^ socialSecurity + federalTaxes + stateTaxes! !

!Withholder methodsFor: 'as yet unclassified' stamp: 'MattMomont 6/22/2014 20:46'!
withholdingOn: aCheck for: anEmployee at: aDate
	self subclassResponsibility.! !
Employee initialize!
ValueWithHistory initialize!
